class PermSet:
    def __init__(self, rn: str, action: str):
        self.rn, self.action = self.__normalize(rn, {action})
        self.action = self.action.pop()

    def __normalize(self, rn: str, actions: set):
        assert type(rn) is str
        assert type(actions) is set

        rn = [p.strip().lower() for p in rn.split(":")]
        actions = {a.strip().lower() for a in actions}

        return rn, actions

    def __cmp_actions(self, actions):
        return "*" in actions or self.action in actions

    def __cmp_rn(self, rn):
        # e.g. rn:a:b >>= rn:a:b
        if self.rn == rn:
            return True

        # e.g. rn:a:b:* >>= rn:a:b
        if self.rn + ["*"] == rn:
            return True

        # e.g. rn:a:* >>= rn:a:b:c
        return rn[-1] == "*" and rn[:-1] == self.rn[:len(rn)-1]

    def check(self, rn: str, actions: set):
        rn, actions = self.__normalize(rn, actions)

        return self.__cmp_actions(actions) and self.__cmp_rn(rn)



import os
import requests as req
from .env import Env
from .auth.keycloak import Keycloak


def load_configs(role):
    env = Env()

    configs_host = env("CONFIGS_HOST", default="configs.avd.al")
    configs_client = env("CONFIGS_CLIENT_ID", default="")
    configs_secret = env("CONFIGS_CLIENT_SECRET", default="")

    iam_host = env("IAM_HOST", default="iam.avd.al")
    iam_realm = env("IAM_REALM", default="groot")

    if not configs_client:
        print("CONFIGS_CLIENT_ID not set. Skipping remote configs")
        return

    keycloak = Keycloak(iam_host, iam_realm, configs_client, configs_secret)
    token, error = keycloak.access_token()

    if not token:
        print("failed to get a token due to", error)
        return

    def load_role(role):
        res = req.get(f"https://{configs_host}/api/v1/roles/{role}/configs", headers={
            "Authorization": f"Bearer {token}",
        })

        if not res.ok:
            return

        for k, v in res.json().items():
            os.environ[k] = str(v)

    load_role(role)
    load_role("common")



import os
import re
from . import annotations

_envre = re.compile(r'''^(?:export\s*)?([_a-zA-Z][\w_]*)\s*=\s*(.*)$''')
_varre = re.compile(r'''\$([\w_]+)''')
_include_re = re.compile(r'''^#include\s+(.*)\s*$''')


def load_env(env_file: str):
    env_file = os.path.abspath(env_file)
    envs = getattr(os.environ, "__env_files", set())

    if env_file in envs:
        return

    envs.add(env_file)
    os.environ.__env_files = envs

    if not os.path.isfile(env_file):
        return

    with open(env_file, "r") as f:
        for line in f.readlines():
            match = _include_re.match(line)
            if match is not None:
                file = match.group(1).strip()
                load_env(file)

            match = _envre.match(line)
            if match is not None:
                key = match.group(1)
                value = match.group(2).strip('"').strip("'")
                for var in _varre.findall(value):
                    value = value.replace(f"${var}", os.environ.get(var, ""))

                os.environ[key] = value


def load_appname(default=None):
    return os.environ.get("APP_NAME", default or "UNNAMED_FLASK_APP").upper()


class Env:
    def __init__(self, prefix=None):
        self.prefix = prefix

    @annotations.enforce_types
    def __call__(self, name: str, default: any = None, cast: type = str, prefixed: bool = True):
        if prefixed and self.prefix:
            name = f"{self.prefix}_{name}"

        value = os.environ.get(name, default=default)

        if value is None:
            raise Exception(f"{name} not found. Declare it as environment variable or provide a default value.")

        if cast is not str:
            try:
                value = cast(value)
            except ValueError:
                raise Exception(f"cannot cast '{value}' into {cast.__name__}")

        return value



import os


class Error(Exception):
    def __init__(self, message="something went wrong", code=500):
        Exception.__init__(self)

        self.message = message
        self.code = code

        assert(self.code > 399 and self.code < 600)

    def _debugging(self):
        debug = os.environ.get("APP_DEBUG")

        return debug and (debug.lower() == "true" or debug == "1")

    def json(self):
        msg = "something went wrong" if self.code > 499 and not self._debugging() else self.message

        body = {
            "code": self.code,
        }

        if type(msg) is dict:
            for k, v in msg.items():
                body[k] = v
        else:
            body["message"] = msg

        return body


class NotFound(Error):
    def __init__(self, resource):
        super(NotFound, self).__init__(f"{resource}: not found", 404)


class Conflict(Error):
    def __init__(self, resource):
        super(Conflict, self).__init__(f"{resource}: already exists", 409)


class BadRequest(Error):
    def __init__(self, message):
        super(BadRequest, self).__init__(message, 400)


class MethodNotAllowed(Error):
    def __init__(self, method):
        super(MethodNotAllowed, self).__init__(f"{method} not allowed", 405)


class Unauthorized(Error):
    def __init__(self, scope):
        super(Unauthorized, self).__init__(f"unauthorized: {scope} scope is required", 401)



import inspect
import typing as t
from inspect import Parameter


def argmap(f):
    def wrapper(*args, **kwargs):
        sig = inspect.signature(f)

        checks = {}
        sig = inspect.signature(f)
        i = 0
        argc = len(args)
        for k, v in sig.parameters.items():
            if k in kwargs:
                checks[k] = (v.annotation, kwargs[k])
            elif argc > 0:
                checks[k] = (v.annotation, args[i])
                i += 1
                argc -= 1
            elif v.default is not None:
                checks[k] = (v.annotation, v.default)

        return checks

    return wrapper


def enforce_types(f: t.Callable) -> t.Callable:
    def wrapper(*args, **kwargs):
        checks = argmap(f)(*args, **kwargs)

        for arg, (annotation, value) in checks.items():
            if annotation == Parameter.empty:
                continue

            if annotation is not any and annotation != type(value):
                fn = f.__name__
                e = annotation.__name__
                a = type(value).__name__

                raise ValueError(f"{fn}: param {arg} expects a value of type {e}, got {a}")

        return f(*args, **kwargs)

    wrapper.__name__ = f.__name__
    return wrapper


def auto_attrs(f: t.Callable) -> t.Callable:
    def wrapper(*args, **kwargs):
        if f.__name__ != "__init__":
            return

        map = argmap(f)(*args, **kwargs)

        if not "self" in map:
            return

        _, self = map["self"]
        del map["self"]

        for arg, (_, value) in map.items():
            setattr(self, arg, value)

        return f(*args, **kwargs)

    wrapper.__name__ = f.__name__
    return wrapper



import unittest
from avdal import annotations
from avdal import rbac


def fails(f, *args, **kwargs):
    try:
        f(*args, **kwargs)
        return False
    except:
        return True


class TestTyping(unittest.TestCase):
    def test_enforce_static_annotations(self):

        @annotations.enforce_types
        def f1(a: str, b: int):
            pass

        self.assertFalse(fails(f1, "a", 1))
        self.assertTrue(fails(f1, "a", "1"))

        @annotations.enforce_types
        def f2(a: str, b: int, c: str = "", d=1):
            pass

        self.assertFalse(fails(f2, "a", 1, "", 1))
        self.assertFalse(fails(f2, "a", 1, "", ""))
        self.assertTrue(fails(f2, "a", 1, 1, ""))

        @annotations.enforce_types
        def f3(a=1, b=2, c: str = ""):
            pass

        self.assertFalse(fails(f3, a=2))
        self.assertFalse(fails(f3, 0, b=2))
        self.assertFalse(fails(f3, c="1", a=2))
        self.assertFalse(fails(f3, 1, 2))
        self.assertTrue(fails(f3, 1, 2, 1))

    def test_auto_attrs(self):
        class A:
            @annotations.auto_attrs
            def __init__(self, a, b, c, d=4, e=5, f=6):
                pass

        a = A(1, 2, 3, 7, f=8)

        self.assertTrue(getattr(a, "a"), 1)
        self.assertTrue(getattr(a, "b"), 2)
        self.assertTrue(getattr(a, "c"), 3)
        self.assertTrue(getattr(a, "d"), 7)
        self.assertTrue(getattr(a, "e"), 5)
        self.assertTrue(getattr(a, "f"), 8)


class TestRbac(unittest.TestCase):
    def test_permset(self):
        r1 = rbac.PermSet("rn:a:b", "read")

        self.assertTrue(r1.check("rn:a:b", {"read"}))
        self.assertTrue(r1.check("rn:a:b", {"*"}))
        self.assertFalse(r1.check("rn:a:b", {"write"}))
        self.assertFalse(r1.check("rn:a:b:c", {"read"}))
        self.assertFalse(r1.check("rn:a:b:c", {"write"}))
        self.assertFalse(r1.check("rn:a", {"write"}))
        self.assertFalse(r1.check("rn:a:*", {"write"}))
        self.assertTrue(r1.check("rn:a:*", {"read"}))
        self.assertTrue(r1.check("rn:*", {"read"}))
        self.assertTrue(r1.check("*", {"read"}))
        self.assertFalse(r1.check("*", {"write"}))
        self.assertTrue(r1.check("*", {"*"}))


        r2 = rbac.PermSet("rn:a:b", "write")

        self.assertFalse(r2.check("rn:a:b", {"read"}))
        self.assertTrue(r2.check("rn:a:b", {"write"}))
        self.assertTrue(r2.check("rn:a:b", {"write", "read"}))
        self.assertTrue(r2.check("rn:a:b", {"update", "write", "read"}))
        self.assertFalse(r2.check("rn:a:b", {"update"}))
        self.assertFalse(r2.check("rn:a:b", set()))
        self.assertTrue(r2.check("rn:a:b", {"*"}))
        self.assertTrue(r2.check("rn:a:b", {"*", "read"}))
        self.assertTrue(r2.check("rn:a:b", {"*", "write"}))
        self.assertTrue(r2.check("rn:a:b", {"*", "update", "read"}))
        self.assertFalse(r2.check("rn:a:b:c", {"*", "update", "read"}))
        self.assertTrue(r2.check("rn:a:b:*", {"*", "update", "read"}))

        r3 = rbac.PermSet("rn:a:b", "read")
        self.assertTrue(r3.check("rn:a:*", {"read"}))
        self.assertTrue(r3.check("rn:a:b:*", {"read"}))
        self.assertFalse(r3.check("rn:a:b:c", {"read"}))
        self.assertFalse(r3.check("rn:a:c", {"read"}))
        self.assertFalse(r3.check("rn:a", {"read"}))




import requests


class Keycloak:
    def __init__(self, host, realm, client_id, client_secret):
        self.client_id = client_id
        self.client_secret = client_secret

        self.token_endpoint = f"https://{host}/auth/realms/{realm}/protocol/openid-connect/token"

    def access_token(self):
        h = {
            "Content-Type": "application/x-www-form-urlencoded"
        }

        d = f"grant_type=client_credentials&client_id={self.client_id}&client_secret={self.client_secret}"
        response = requests.request("POST", self.token_endpoint, headers=h, data=d)

        if not response.ok:
            return None, response.json()

        return response.json()["access_token"], None
